# importiamo le librerie necessarie per l'hash e le altre funzioni di base
from hashlib import sha256
import json
import time

#importiamo le librerie necessarie di Flask
from flask import Flask, request
import requests


class Block:
    # Costruttore per la classe Block
    # param index: ID Univoco del blocco
    # param transactions: lista delle transazioni
    # param timestamp: data di creazione del blocco
    # param previous_hash: hash del blocco precedente
    # param nonce: parametro necessario per la PoW
    def __init__(self, index, transactions, timestamp, previous_hash, nonce=0):
        self.index = index
        self.transactions = transactions
        self.timestamp = timestamp
        self.previous_hash = previous_hash
        self.nonce = nonce

    # Funzione che ritorna l'hash del contenuto del blocco
    def compute_hash(self):
        block_string = json.dumps(self.__dict__, sort_keys=True)
        return sha256(block_string.encode()).hexdigest()


class Blockchain:
    # Livello di difficolt dell'algoritmo PoW
    difficulty = 2

    # Costruttore della classe Blockchain
    def __init__(self):
        self.unconfirmed_transactions = [] #lista delle transazioni non confermate
        self.chain = []

    # La funzione genera il blocco di genesi e lo aggiunge alla blockchain.
    # Il primo blocco ha un indice pari a 0,
    # previous_hash uguale a 0  un valore hash valido.
    def create_genesis_block(self):
        genesis_block = Block(0, [], 0, "0")
        genesis_block.hash = genesis_block.compute_hash()
        self.chain.append(genesis_block)

    # Una semplice propriet che prende l'ultimo blocco della catena.
    # P.S. Nella catena c' sempre un primo blocco che sar il blocco di genesi
    @property
    def last_block(self):
        return self.chain[-1]

    # Aggiunge un blocco alla catena dopo la verifica. La verifica include che
    # - la proof-of-work sia corretta
    # - l'hash precedente sia corretto e corrisponda al valore hash del nuovo blocco nella catena
    def add_block(self, block, proof):
        
        previous_hash = self.last_block.hash # prende l'hash dell'ultimo blocco nella catena

        # se l'hash precedente non  uguale al previuos_hash del
        # blocco che sto per aggiungere, ritorna falso
        if previous_hash != block.previous_hash:
            return False

        # controlla anche la validit della proof del blocco
        if not Blockchain.is_valid_proof(block, proof):
            return False

        block.hash = proof
        self.chain.append(block)
        return True

    # La funzione verifica diversi valori di nonce fino
    # a quando non trova il valore hash soddisfacente.
    @staticmethod
    def proof_of_work(block):
        block.nonce = 0

        computed_hash = block.compute_hash()
        while not computed_hash.startswith('0' * Blockchain.difficulty):
            block.nonce += 1
            computed_hash = block.compute_hash()

        return computed_hash

    # Funzione che aggiunge della transazione all'array delle
    # transazioni non confermate
    def add_new_transaction(self, transaction):
        self.unconfirmed_transactions.append(transaction)

    # Controlla se block_hash  il valore hash valido del blocco
    # e se soddisfa i criteri di difficolt
    @classmethod
    def is_valid_proof(cls, block, block_hash):
        return (block_hash.startswith('0' * Blockchain.difficulty) and
                block_hash == block.compute_hash())

    # funzione che controlla se l'interna blockchain  corretta
    @classmethod
    def check_chain_validity(cls, chain):
        result = True
        previous_hash = "0"

        # il ciclo for viene eseguito per tutti i blocchi presenti nella catena
        for block in chain:
            block_hash = block.hash
            
            # cancella il campo hash per ricalcolare il valore hash usando la funzione compute_hash
            delattr(block, "hash")

            if not cls.is_valid_proof(block, block_hash) or previous_hash != block.previous_hash:
                result = False
                break

            block.hash, previous_hash = block_hash, block_hash

        return result

    # La funzione aggiunge le transazioni pendenti dentro la blockchain
    # aggiungendole al blocco e rilasciando il PoW
    def mine(self):
        if not self.unconfirmed_transactions:
            return False # se non ci sono transazioni pendenti, ritorna falso

        last_block = self.last_block # recupera l'ultimo blocco

        # - crea un nuovo blocco, assegnando un nuovo indice sulla base dell'indice dell'ultimo blocco
        #   incrementato di una unit
        # - imposta dentro il blocco tutte le transazioni pendenti (array)
        # - imposta come previous_hash, l'hash dell'ultimo blocco (last_block)
        new_block = Block(index=last_block.index + 1,
                          transactions=self.unconfirmed_transactions,
                          timestamp=time.time(),
                          previous_hash=last_block.hash)

        proof = self.proof_of_work(new_block) # calcola la proof-of-work del nuovo blocco
        self.add_block(new_block, proof) # aggiunge il nuovo blocco alla blockchain
        self.unconfirmed_transactions = [] # svuota l'array delle transazioni pendenti
        return True

# Creo l'istanze dell'applicazione Flask passando __name__come argomento
app = Flask(__name__)

# Inizializzo la blockchain (con il relativo il blocco di genesi)
blockchain = Blockchain()
blockchain.create_genesis_block()

# Impostiamo innanzitutto un set vuoto che conterr l'indirizzo host degli altri nodi della rete
peers = set()

# endpoint to submit a new transaction. This will be used by
# our application to add new data (posts) to the blockchain
@app.route('/new_transaction', methods=['POST'])
def new_transaction():
    tx_data = request.get_json() #dati in arrivo dal form attraverso il submit
    required_fields = ["author", "content"]

    # controllo tutti i campi obbligatori
    for field in required_fields:
        if not tx_data.get(field):
            # ritorna un errore 404 nel caso uno dei due campi fosse vuoto
            return "Transazione invalida", 404

    tx_data["timestamp"] = time.time() # aggiunge la data in formato timestamp
    blockchain.add_new_transaction(tx_data) # aggiunge la transazione nell'array delle transazioni pendenti
    return "Success", 201

# endpoint che restituisce una copia della catena sottoforma
# di stringa JSON. Questo endpoint potr essere usato per interrogare
# tutti i blocchi e le transazioni presenti nella blockchain.
@app.route('/chain', methods=['GET'])
def get_chain():
    chain_data = []
    for block in blockchain.chain:
        chain_data.append(block.__dict__)
    return json.dumps({"length": len(chain_data),
                       "chain": chain_data,
                       "peers": list(peers)})


# endpoint che permette di inviare una richiesta di mining
# convalidando le transazioni non verificate (se ovviamente
# l'array ha delle transazioni pendenti)
@app.route('/mine', methods=['GET'])
def mine_unconfirmed_transactions():
    result = blockchain.mine()
    if not result:
        return "Nessuna transazione da minare"
    else:
        # Assicuriamoci di avere la lunghezza massima della catena,
        # prima di comunicare con la rete
        chain_length = len(blockchain.chain)
        consensus()
        if chain_length == len(blockchain.chain):
            # annuncia a tutti i nodi peer che il blocco  stato minato
            announce_new_block(blockchain.last_block)
        return "Block #{} minato.".format(blockchain.last_block.index)


# Endpoint per aggiungere un nuovo peer alla rete
@app.route('/register_node', methods=['POST'])
def register_new_peers():
    # Recupera l'indirizzo dell'host del nodo peer
    node_address = request.get_json()["node_address"]

    # nel caso in cui non viene fornito l'indirizzo, ritorna un errore
    if not node_address:
        return "Invalid data", 400

    # Aggiungi l'indirizzo del nodo al set
    peers.add(node_address)

    # Ritorna la blockchain
    return get_chain()

# Endpoint per registrare un nuovo nodo. Al suo interno
# effettua una chiamata alla rotta "register_node" per
# registrare il nodo corrente con il nodo remoto specificato
# nella richiesta POST e alla fine aggiorna la rete blockchain
@app.route('/register_with', methods=['POST'])
def register_with_existing_node():
    
    node_address = request.get_json()["node_address"]
    if not node_address:
        return "Dati non validi", 400

    data = {"node_address": request.host_url}
    headers = {'Content-Type': "application/json"}

    # Richiesta di registrazione con il nodo remoto
    response = requests.post(node_address + "/register_node",
                             data=json.dumps(data), headers=headers)

    if response.status_code == 200:
        global blockchain
        global peers

        # aggiorna la catena e i peers
        chain_dump = response.json()['chain']
        blockchain = create_chain_from_dump(chain_dump)
        peers.update(response.json()['peers'])
        return "Registrazione avvenuta", 200
    else:
        # Se viene sollevato un errore, l'API ritorna un response
        return response.content, response.status_code

    return "ok", 200

def create_chain_from_dump(chain_dump):
    generated_blockchain = Blockchain()
    generated_blockchain.create_genesis_block()
    for idx, block_data in enumerate(chain_dump):
        if idx == 0:
            continue  # salta il blocco di genesi
        block = Block(block_data["index"],
                      block_data["transactions"],
                      block_data["timestamp"],
                      block_data["previous_hash"],
                      block_data["nonce"])
        proof = block_data['hash']
        added = generated_blockchain.add_block(block, proof)
        if not added:
            raise Exception("Catena manomessa!")
    return generated_blockchain


# Endpoint che aggiunge il blocco appena minato alla catena.
# Il blocco richiede di essere verificato prima di poter
# essere inserito alla catena.
@app.route('/add_block', methods=['POST'])
def verify_and_add_block():
    block_data = request.get_json()
    block = Block(block_data["index"],
                  block_data["transactions"],
                  block_data["timestamp"],
                  block_data["previous_hash"],
                  block_data["nonce"])

    proof = block_data['hash']
    added = blockchain.add_block(block, proof)

    if not added:
        return "Blocco scartato dal nodo", 400

    return "Blocco aggiunto alla catena", 201


# Endpoint che ritorna le transazioni attualmente in stato di pending
@app.route('/pending_tx')
def get_pending_tx():
    return json.dumps(blockchain.unconfirmed_transactions)

# La funzione implementa un semplice algoritmo di consenso.
# Va a controllare la lunghezza della catena. Se la lunghezza trovata  maggiore di
# tutte quelle possibili, allora la catena viene sostituita.
def consensus():
    global blockchain

    longest_chain = None
    current_len = len(blockchain.chain)

    for node in peers:
        response = requests.get('{}chain'.format(node))
        length = response.json()['length']
        chain = response.json()['chain']
        if length > current_len and blockchain.check_chain_validity(chain):
            # trovata una catena di lunghezza superiore e viene sostituita
            current_len = length
            longest_chain = chain

    if longest_chain:
        blockchain = longest_chain
        return True

    return False

# Funzione che informa la rete dopo aver minato un blocco.
# Gli altri nodi peer devono solo verificare il PoW e
# aggiungere la stringa corrispondente.
def announce_new_block(block):
    for peer in peers:
        url = "{}add_block".format(peer)
        headers = {'Content-Type': "application/json"}
        requests.post(url,
                      data=json.dumps(block.__dict__, sort_keys=True),
                      headers=headers)


